# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from .._unit import Unit
from .decorators import cost

__all__ = ('MagneticSeparator',)

@cost('Flow rate', units='kg/hr', CE=576, cost=533471, S=333333, BM=4.16, n=0.6)
class MagneticSeparator(Unit): 
    _N_outs = 2
    def _run(self):
        self.outs[0].copy_like(self.ins[0])


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
__version__ = '2.31.23'

#: dict[str, float] Price of stream utilities (in USD/kg) which are defined as 
#: inlets and outlets to unit operations.
stream_utility_prices = {'Natural gas': 0.218,
                         'Ash disposal': -0.0318}

#: Chemical engineering plant cost index (defaults to 567.5 at 2017).
CE = 567.5 

#: Whether to label the ID of streams with sources and sinks in process 
#: flow diagrams.
LABEL_PROCESS_STREAMS_IN_DIAGRAMS = True

#: Whether to automatically generate diagrams when displaying an object in the
#: IPython console.
ALWAYS_DISPLAY_DIAGRAMS = True

#: Whether to ignore unit graphics and display unit nodes as dots in process
#: flow diagrams.
MINIMAL_UNIT_DIAGRAMS = False

#: Whether to number unit operations in diagrams according to their order in the system path.
LABEL_PATH_NUMBER_IN_DIAGRAMS = False

#: Whether to clock the simulation time of unit operations in diagrams.
PROFILE_UNITS_IN_DIAGRAMS = False

#: Whether to raise exception regarding problems displaying graphviz diagrams
RAISE_GRAPHVIZ_EXCEPTION = False

# %% Workaround for readthedocs, which fails to cache numba

import numba
try:
    @numba.njit(cache=True)
    def f_dummy(): pass
except RuntimeError:
    def njit(*args, **kwargs):
        kwargs['cache'] = False
        return numba.jit(*args, **kwargs)
    numba.njit = njit
    del njit
else:
    del f_dummy

# %% Initialize BioSTEAM 

from thermosteam import Chemical, Chemicals, Thermo, Stream, MultiStream, settings, speed_up
from ._heat_utility import HeatUtility, UtilityAgent
from ._power_utility import PowerUtility
from ._unit import Unit
from . import _system
from ._system import *
from . import process_tools
from .process_tools import *
from . import _tea
from ._tea import *
from . import _flowsheet
from ._flowsheet import *
from . import utils
from . import units
from .units import *
from . import evaluation
from .evaluation import *
from . import exceptions
from . import report

__all__ = (
    'Unit', 'PowerUtility', 'HeatUtility', 'UtilityAgent',
    'utils', 'units', 'evaluation', 'Chemical', 'Chemicals', 'Stream',
    'MultiStream', 'settings', 'exceptions', 'report',
    'process_tools', *_system.__all__, *_flowsheet.__all__, 
    *_tea.__all__, *units.__all__, *evaluation.__all__, 
    *process_tools.__all__, 
)


# -*- coding: utf-8 -*-
"""
Created on Sun Apr 26 13:54:54 2020

@author: yoelr
"""
from numba import njit

__all__ = ('approx2step',)

@njit(cache=True)
def approx2step(val, x0, dx):
    """Approximate value, val, to closest increment/step, dx, starting from x0."""
    while True:
        if x0 > val: break
        x0 += dx
    return x0



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from . import Facility
from ..decorators import cost
from ... import HeatUtility
import numpy as np

__all__ = ('RefrigerationPackage',)

class RefrigerationPackage(Facility):
    """
    Create a refrigeration package that converts refrigeration usage to electricity demand [1]_.
    
    Parameters
    ----------
    ID : str, optional
        Unit ID.
    
    References
    ----------
    .. [1] Seider, Warren D., et al. (2017). "Cost Accounting and Capital Cost
        Estimation". In Product and Process Design Principles: Synthesis,
        Analysis, and Evaluation (pp. 450-455). New York: Wiley.
    
    """
    ticket_name = 'RP'
    network_priority = 0
    _N_heat_utilities = 1
    _units = {'Duty': 'kJ/hr'}
    def __init__(self, ID='', agent=None, efficiency=0.35):
        raise NotImplementedError('not yet ready for users')
        self.agent = agent or HeatUtility.get_cooling_agent('propane')
        self.efficiency = efficiency
        super().__init__(ID)
        
    def _load_refrigeration_utilities(self):
        self.refrigeration_utilities = ru = set()
        ID = self.agent.ID
        for u in self.other_units:
            if u is self: continue
            for hu in u.heat_utilities:
                agent = hu.agent 
                if agent and agent.ID == ID: ru.add(hu)
        
    def _design(self):
        self._load_refrigeration_utilities()
        ru = self.refrigeration_utilities
        self.design_results['Duty'] = duty = sum([i.duty for i in ru])        
        hu, = self.heat_utilities
        hu.mix_from(ru)
        hu.reverse()
        self.power_utility.consumption = duty /3600 * self.efficiency
        


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from collections import namedtuple

__all__ = ('StreamLinkOptions',
           'static',
           'static_flow_and_phase',
           'static_link_options',
           'static_flow_link_options',
           'static_flow_and_phase_options')

# %% Linking options

StreamLinkOptions = namedtuple('StreamLinkOptions', ('flow', 'phase', 'TP'), module=__name__)
static_link_options = StreamLinkOptions(flow=True, TP=True, phase=True)
static_flow_and_phase_options = StreamLinkOptions(flow=True, TP=False, phase=True)
static_flow_link_options = StreamLinkOptions(flow=True, TP=False, phase=False)

def _run_static(self):
    self._outs[0].copy_like(self._ins[0])

def static(cls):
    cls._stream_link_options = static_link_options
    cls._N_ins = cls._N_outs = 1
    cls._run = _run_static
    return cls

def _run_static_flow(self):
    outlet = self._outs[0]
    inlet = self._ins[0]
    outlet.phases = inlet.phases
    outlet.copy_flow(inlet)

def static_flow_and_phase(cls):
    cls._stream_link_options = static_flow_and_phase_options
    cls._N_ins = cls._N_outs = 1
    return cls

del namedtuple


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
This module includes arbitrary classes and functions.

"""
import biosteam as bst
from thermosteam import Chemicals
from math import ceil

__all__ = ('factor', 'checkbounds', 'strtuple',
           'format_title', 'format_unit_name',
           'remove_undefined_chemicals',
           'default_chemical_dict', 'subgroup',
           'repr_subgroups', 'repr_items')

# %% Number functions

def factor(base_units, new_units):
    if base_units == new_units: return 1
    else: return bst._Q(1, base_units).to(new_units).magnitude

def checkbounds(x, bounds):
    return bounds[0] < x < bounds[1]

# %% String functions

def strtuple(iterable):
    """Return string of all items in the tuple""" 
    string = ''
    function = type(strtuple)
    for i in iterable:
        if isinstance(i , function):
            string += i.__name__ + ', '
        else:
            string += str(i) + ', '
    string = string.rstrip(', ')
    string = '(' + string + ')'
    return string
        
def format_title(line):
    line = line.replace('_', ' ')
    words = []
    word = ''
    last = ''
    for i in line:
        if i.isupper() and last.isalpha() and not last.isupper():
            words.append(word)
            word = i
        else:
            word += i
        last = i
    words.append(word)
    line = ''
    for word in words:
        N_letters = len(word)
        if N_letters > 1:
            line += word + ' '
        else:
            line += word
    line = line.strip(' ')
    first_word, *rest = line.split(' ')
    words = [first_word[0].capitalize() + first_word[1:]]
    for word in rest:
        if not all([(i.isupper() or not last.isalpha()) for i in word]):
            word = word.lower()
        words.append(word)
    return ' '.join(words)

def format_unit_name(name):
    return ''.join([i[0].capitalize() + i[1:] for i in name.split(' ')])
    
def subgroup(items, size=5):
    return [items[size*i: size*(i+1)] for i in range(int(ceil(len(items) / size)))]

def repr_subgroups(subgroups):
    return [', '.join([str(i) for i in j]) for j in subgroups]

def repr_items(start, items, subgroup_size=5, brackets=None):
    N_spaces = len(start)
    if brackets:
        left, right = brackets
        N_spaces += 1
    else:
        left = ''; right = ''
    subgroups = repr_subgroups(subgroup(items, subgroup_size))
    dlim = ",\n" + " " * N_spaces
    return start + left + dlim.join(subgroups) + right

# %% Chemical management

def remove_undefined_chemicals(data: dict, chemicals: Chemicals):
    for i in tuple(data):
        if i not in chemicals: del data[i]

def default_chemical_dict(dct, chemicals, g, l, s, n=None):
    if n is None: n = l
    for i in chemicals:
        ID = i.ID
        if ID not in dct:
            locked_state = i.locked_state
            if locked_state == 'g': dct[ID] = g
            elif locked_state == 'l': dct[ID] = l
            elif locked_state == 's': dct[ID] = s
            elif locked_state is None: dct[ID] = n
            else: raise RuntimeError(f"unknown locked state '{locked_state}' of chemical '{i}'")



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
This module contains unit operations for mixing.

.. contents:: :local:
    
Unit operations
---------------
.. autoclass:: biosteam.units.mixing.Mixer
.. autoclass:: biosteam.units.mixing.SteamMixer

"""
from ..utils import InletPort, OutletPort, ignore_docking_warnings
from .._unit import Unit
from .._graphics import mixer_graphics
import flexsolve as flx
import biosteam as bst

__all__ = ('Mixer', 'SteamMixer', 'FakeMixer', 'MockMixer')

class Mixer(Unit):
    """
    Create a mixer that mixes any number of streams together.
    
    Parameters
    ----------
    ins : streams
        Inlet fluids to be mixed.
    outs : stream
        Mixed outlet fluid.
    Examples
    --------
    Mix two streams:
    
    >>> from biosteam import units, settings, Stream
    >>> settings.set_thermo(['Ethanol', 'Water'], cache=True)
    >>> s1 = Stream('s1', Water=20, T=350)
    >>> s2 = Stream('s2', Ethanol=30, T=300)
    >>> M1 = units.Mixer('M1', ins=(s1, s2), outs='s3')
    >>> M1.simulate()
    >>> M1.show()
    Mixer: M1
    ins...
    [0] s1
        phase: 'l', T: 350 K, P: 101325 Pa
        flow (kmol/hr): Water  20
    [1] s2
        phase: 'l', T: 300 K, P: 101325 Pa
        flow (kmol/hr): Ethanol  30
    outs...
    [0] s3
        phase: 'l', T: 315.11 K, P: 101325 Pa
        flow (kmol/hr): Ethanol  30
                        Water    20
    
    
    """
    _graphics = mixer_graphics
    _N_outs = 1
    _N_ins = 2
    _ins_size_is_fixed = False
    
    def _assert_compatible_property_package(self): 
        pass # Not necessary for mixing streams
    
    def _run(self):
        s_out, = self.outs
        s_out.mix_from(self.ins)
        
    @ignore_docking_warnings
    def insert(self, stream):
        """
        Insert Mixer object between two units at a given stream connection.
        
        Examples
        --------
        >>> from biosteam import *
        >>> settings.set_thermo(['Water'], cache=True)
        >>> feed = Stream('feed')
        >>> other_feed = Stream('other_feed')
        >>> P1 = Pump('P1', feed, 'pump_outlet')
        >>> H1 = HXutility('H1', P1-0, T=310)
        >>> M1 = Mixer('M1', other_feed, 'mixer_outlet')
        >>> M1.insert(P1-0)
        >>> M1.show()
        Mixer: M1
        ins...
        [0] other_feed
            phase: 'l', T: 298.15 K, P: 101325 Pa
            flow: 0
        [1] pump_outlet  from  Pump-P1
            phase: 'l', T: 298.15 K, P: 101325 Pa
            flow: 0
        outs...
        [0] mixer_outlet  to  HXutility-H1
            phase: 'l', T: 298.15 K, P: 101325 Pa
            flow: 0
        
        """
        sink = stream.sink
        sink.ins.replace(stream, self.outs[0])
        self.ins.append(stream)
        

class SteamMixer(Unit):
    """
    Create a mixer that varies the flow of steam to achieve a specified outlet
    pressure and varies the flow of process water to achieve a specified 
    solids loading (by wt).
    
    Parameters
    ----------
    ins : stream sequence
        [0] Feed    
        [1] Steam
        [2] Process water
    outs : stream
        Mixed product.
    P : float
        Outlet pressure.
    
    """
    _N_outs = 1
    _N_ins = 3
    _ins_size_is_fixed = False
    _N_heat_utilities = 1
    _graphics = mixer_graphics
    installation_cost = purchase_cost = 0.
    def __init__(self, ID='', ins=None, outs=(), thermo=None, *, 
                 P, solids_loading=None, liquid_IDs=['7732-18-5']):
        Unit.__init__(self, ID, ins, outs, thermo)
        self.P = P
        self.solids_loading = solids_loading
        self.liquid_IDs = tuple(liquid_IDs)
    
    @property
    def steam(self):
        return self.ins[1]
    
    def reset_cache(self): 
        for utility in bst.HeatUtility.heating_agents:
            if utility.P > self.P: break
        self.steam.copy_like(utility)
    
    def pressure_objective_function(self, steam_mol):
        try:
            feed, steam, process_water, *others = self.ins
        except ValueError:
            feed, steam, *others = self.ins
        feeds = [feed, *others]
        mixed = self.outs[0]
        steam.imol[self.liquid_IDs] = steam_mol
        solids_loading = self.solids_loading
        if solids_loading is not None:
            chemicals = self.chemicals
            index = chemicals.get_index(self.liquid_IDs)
            F_mass_feed = sum([i.F_mass for i in feeds if i])
            available_water = (18.01528 * sum([i.mol[index].sum() for i in feeds if i])).sum()
            required_water = (F_mass_feed - available_water) * (1. - solids_loading) / solids_loading
            try:
                process_water.imol['7732-18-5'] = max(required_water - available_water, 0.) / 18.01528
            except NameError:
                raise RuntimeError('missing process water stream')
        mixed.mix_from(self.ins)
        P_new = mixed.chemicals.Water.Psat(min(mixed.T, mixed.chemicals.Water.Tc - 1))
        return self.P - P_new
    
    def _setup(self):
        super()._setup()
        if self.steam.isempty(): self.reset_cache()
        self.outs[0].P = self.P
    
    def _run(self):
        steam = self.ins[1]
        steam_mol = steam.F_mol or 1.
        f = self.pressure_objective_function
        steam_mol = flx.IQ_interpolation(f, *flx.find_bracket(f, 0., steam_mol, None, None), 
                                         xtol=1e-2, ytol=1e-4, 
                                         maxiter=500, checkroot=False)
        
    def _design(self): 
        steam = self.ins[1]
        mixed = self.outs[0]
        self.heat_utilities[0](steam.H, mixed.T)
        
class MockMixer(Unit):
    """
    Create a MockMixer object that does nothing when simulated.
    """
    _graphics = Mixer._graphics
    _N_ins = 2
    _N_outs = 1
    _ins_size_is_fixed = False
    
    def _run(self): pass

MockMixer.line = 'Mixer'
FakeMixer = MockMixer    


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from .. import Unit
from .decorators import cost
from thermosteam.reaction import Reaction, ParallelReaction

__all__ = ('Transesterification',)

@cost('Volume', 'Reactor',
      CE=525.4, cost=15000, n=0.55, kW=1.5, BM=4.3,)
class Transesterification(Unit):
    """
    Create a transesterification reactor that converts 'TAG', 'DAG', 'MAG' and 'Methanol'
    to 'Biodiesel' and 'Glycerol'. Finds the amount of catalyst 'NaOCH3'
    required and consumes it to produce 'NaOH' and 'Methanol'.
    
    Parameters
    ----------
    ins : stream sequence
        * [0] Lipid feed
        * [1] Methanol feed (includes catalyst)
    outs : stream
        Reactor effluent.
    efficiency : float
        Efficiency of conversion (on a 'Lipid' basis).
    excess_methanol : float
        Excess methanol fed in addition to the required stoichiometric amount.
    x_catalyst : float
        Catalyst molar fraction in methanol feed.
    T : float
        Operating temperature [K].
    
    """
    _bounds = {'Volume': (0.1, 20)}
    _units = {'Volume': 'm^3'}
    _tau = 1
    _N_ins = 2
    _N_outs = 1
    _N_heat_utilities = 1

    def _get_design_info(self):
        return (('Residence time', self.tau, 'hr'),
                ('Conversion efficiency', self.efficiency, ''),
                ('Working volume fraction', 0.8, ''))

    def __init__(self, ID='', ins=None, outs=(), *,
                 efficiency, excess_methanol, T, x_catalyst):
        Unit.__init__(self, ID, ins, outs)
        #: [:class:`~thermosteam.ParallelReaction`] Transesterification and catalyst consumption reactions.
        self.transesterification = ParallelReaction([
          Reaction('MAG + Methanol -> Biodiesel + Glycerol', reactant='MAG',  X=efficiency),
          Reaction('DAG + 2Methanol -> 2Biodiesel + Glycerol', reactant='DAG',  X=efficiency),
          Reaction('TAG + 3Methanol -> 3Biodiesel + Glycerol', reactant='TAG',  X=efficiency),
          Reaction('NaOCH3 -> NaOH + Methanol', reactant='NaOCH3', X=1)
        ])
        self.x_catalyst = x_catalyst #: [float] Catalyst molar fraction in methanol feed.
        self.excess_methanol = excess_methanol #: [float] Excess methanol fed in addition to the required stoichiometric amount.
        self.T = T #: [float] Operating temperature (K).
    
    @property
    def tau(self):
        """Residence time (hr)."""
        return self._tau
    @tau.setter
    def tau(self, tau):
        self._tau = tau
    
    @property
    def efficiency(self):
        """Transesterification efficiency."""
        return self.transesterification.X[:-1].mean()
    @efficiency.setter
    def efficiency(self, efficiency):
        self.transesterification.X[:-1] = efficiency
        
    def _run(self):
        feed, methanol = self.ins
        product, = self.outs
        x_catalyst = self.x_catalyst
        x_methanol = 1. - x_catalyst
        required_methanol = 1. + self.excess_methanol
        methanol.imol['Methanol'] = x_methanol * required_methanol * (
              feed.imol['MAG'] + 2 * feed.imol['DAG'] + 3 * feed.imol['TAG'] 
        )
        methanol.imol['NaOCH3'] =  x_catalyst / x_methanol * methanol.imol['Methanol'] 
        product.mix_from([feed, methanol], energy_balance=False)
        self.transesterification(product)
        product.T = self.T
        
    def _design(self):
        effluent = self._outs[0]
        self.design_results['Volume'] = self._tau * effluent.F_vol / 0.8
        self.heat_utilities[0](self.Hnet, effluent.T)

        
    
        
        



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import biosteam as bst
from .splitting import Splitter
from biosteam.units.design_tools import PressureVessel
from math import sqrt, pi

__all__ = ('AdsorptionColumnTSA',)

class AdsorptionColumnTSA(PressureVessel, Splitter):
    """
    Create a temperature swing adsorption (TSA) column. Default parameters
    are heuristic values for adsorption of water and polar components using
    silica gel. The design and cost algorithm of the adsorption column is 
    based on [1]_, [2]_, and [4]_. Three vessels are used: a lead, a guard, 
    and a standby vessel. The lead is in equilibrium, guard holds the 
    breakthrough curve and mass transfer zone, and the standby is in regeneration.
    Once the standby vessel is regenerated, it becomes the guard, the lead becomes
    the standby, and the guard becomes the lead. Therefore, the cycle time is
    the regeneration time.
    
    Parameters
    ----------
    ins : stream sequence
        [0] Feed
        [1] Air
    outs : stream sequence
        [0] Effluent
        [1] Purge
    mean_velocity : float, optional
        Mean velocity of the feed. The diameter of the receiving vessel adjusts
        accordingly. Defaults to 7.2 [m / hr]. Typical velocities are 4 to 14.4 m / hr for liquids [1]_.
    regeneration_air_velocity : float, optional
        Mean velocity of the air used to regenerate the bed. Defaults to 540 [m / hr]. 
        Common velocity range for gasses is 504-2160 m / hr [1]_.
    cycle_time : float, optional
        Time at which the receiving vessel is switched. Defaults to 3 [hr]. 
        Note that 1-2 hours required for thermal-swing-adsorption 
        (TSA) for silica gels. One hr is added to be conservative [2]_. 
    rho_adsorbent : float, optional
        The density of the adsorbent. Defaults to 480 [kg/m3], which is common 
        for silica gels [3]_.
    adsorbent_capacity : float, optional
        Fraction of absorbate that the absorbent can hold. Defaults to 0.1, a 
        conservative heuristic from [2]_.
    T_regeneration : float, optional
        Temperature during the regeneration phase. Defaults to 393.15 [K], which
        is used for silica gels [2]_.
    vessel_material : float, optional
        Vessel material. Defaults to 'Stainless steel 316',
    vessel_type : float, optional
        Vessel type. Defaults to 'Vertical'.
    adsorbate_ID : string
        Name of adsorbate.
    order : tuple[str], optional
        Order of component splits.
    split : dict[str, float] or list[float], optional
        Component splits towards the effluent (0th outlet).
    
    Examples
    --------
    >>> import biosteam as bst
    >>> bst.settings.set_thermo(['Water', 'O2', 'N2', 'Hexane'])
    >>> feed = bst.Stream('feed', Hexane=0.9, Water=0.1, units='kg/hr', total_flow=1000.)
    >>> A1 = bst.AdsorptionColumnTSA('A1', [feed, 'air'], 
    ...     split=dict(Water=0., Hexane=1.0),
    ...     adsorbate_ID='Water',
    ...  )
    >>> A1.simulate()
    >>> A1.results()
    Adsorption column TSA                           Units                   A1
    Low pressure steam  Duty                        kJ/hr             2.75e+05
                        Flow                      kmol/hr                  7.1
                        Cost                       USD/hr                 1.69
    Design              Vessel diameter                                   0.51
                        Vessel length                                     12.6
                        Number of reactors                                   3
                        Vessel type                                   Vertical
                        Length                         ft                 3.83
                        Diameter                       ft                0.156
                        Weight                         lb                 19.9
                        Wall thickness                 in                 0.25
                        Vessel material                    Stainless steel 316
    Purchase cost       Vertical pressure vessel      USD             1.68e+04
                        Platform and ladders          USD                  803
    Total purchase cost                               USD             1.76e+04
    Utility cost                                   USD/hr                 1.69
    
    References
    ----------
    [1] Adsorption basics Alan Gabelman (2017) Adsorption basics Part 1. AICHE
    [2] Seader, J. D., Separation Process Principles: Chemical and Biochemical Operations,” 3rd ed., Wiley, Hoboken, NJ (2011).
    [3] https://www.daisogelusa.com/technical-notes/approximate-packing-density-for-daisogel-bulk-silica-gel/
    [4] Seider, W. D., Lewin,  D. R., Seader, J. D., Widagdo, S., Gani,
        R., & Ng, M. K. (2017). Product and Process Design Principles. Wiley.
        Cost Accounting and Capital Cost Estimation (Chapter 16)
    
    """
    # NOTE: Unit ignores cost of compressing air.
    _N_ins = 2
    _N_outs = 2
    _N_heat_utilities = 1
    
    def __init__(self, 
            ID='', ins=None, outs=(), thermo=None, *,
            mean_velocity=7.2, # m / hr; typical velocities are 4 to 14.4 m /hr for liquids; Adsorption basics Alan Gabelman (2017) Adsorption basics Part 1. AICHE
            regeneration_air_velocity=1332, # Mid point in velocity range for gasses, m / hr; Alan Gabelman (2017) Adsorption basics Part 1. AICHE
            cycle_time=3, # 1-2 hours required for thermal-swing-adsorption (TSA) for silica gels (add 1 hr for conservativeness); Seader, J. D., Separation Process Principles: Chemical and Biochemical Operations,” 3rd ed., Wiley, Hoboken, NJ (2011).
            rho_adsorbent=480, # (in kg/m3) Common for silica gels https://www.daisogelusa.com/technical-notes/approximate-packing-density-for-daisogel-bulk-silica-gel/
            adsorbent_capacity=0.1, # Conservative heuristic from Seider et. al. (2017) Product and Process Design Principles. Wiley
            T_regeneration=120 + 273.15, # For silica gels; Seader, J. D., Separation Process Principles: Chemical and Biochemical Operations,” 3rd ed., Wiley, Hoboken, NJ (2011).
            vessel_material='Stainless steel 316',
            vessel_type='Vertical',
            adsorbate_ID, 
            order=None, 
            split,
        ):
        bst.Splitter.__init__(self, ID, ins, outs, thermo, order=order, split=split)
        self.mean_velocity = mean_velocity
        self.regeneration_air_velocity = regeneration_air_velocity
        self.cycle_time = cycle_time
        self.rho_adsorbent = rho_adsorbent
        self.adsorbent_capacity = adsorbent_capacity
        self.adsorbate_ID = adsorbate_ID
        self.vessel_material = vessel_material
        self.vessel_type = vessel_type
        self.T_regeneration = T_regeneration
        
    @property
    def effluent(self):
        return self.outs[0]
        
    @property
    def regeneration_purge(self):
        return self.outs[1]
    
    def _run(self):
        feed, air = self.ins
        effluent, purge = self.outs 
        feed.split_to(effluent, purge, self.split)
        F_vol_feed = feed.F_vol
        mean_velocity = self.mean_velocity
        self.design_results['Vessel diameter'] = diameter = 2 * sqrt(F_vol_feed / (mean_velocity * pi))
        regeneration_air_velocity = self.regeneration_air_velocity
        diameter = 2 * sqrt(F_vol_feed / (mean_velocity * pi))
        radius = diameter * 0.5
        F_vol_air = radius * radius * regeneration_air_velocity * pi
        purge.T = air.T = self.T_regeneration
        air.P = 10.
        air.imass['N2', 'O2'] = [0.78, 0.32]
        air.phase = 'g'
        air.F_vol = F_vol_air
        purge.mol += air.mol
        purge.phase = 'g'
        
    @property
    def online_time(self):
        return 2 * self.cycle_time # 3 columns (1 quard, 1 active, 1 regenerating)
    
    def _design(self):
        design_results = self.design_results
        feed = self.ins[0]
        rho_adsorbent = self.rho_adsorbent
        adsorbent_capacity = self.adsorbent_capacity
        F_mass_adsorbate = float(feed.imass[self.adsorbate_ID])
        diameter = design_results['Vessel diameter']
        online_length = (
            self.online_time * F_mass_adsorbate / (adsorbent_capacity * rho_adsorbent * diameter)
        ) + 0.61 # equilibrium length plust 2 ft accounting for mass transfer limitations
        design_results['Vessel length'] = length = online_length / 2. # Size of each column
        design_results['Number of reactors'] = 3
        design_results.update(
            self._vessel_design(
                feed.P * 0.000145038, # Pa to psi
                float(diameter) / 3.28084, # m to ft
                float(length) / 3.28084, # m to ft
            )
        )    
        self.heat_utilities[0](self.regeneration_purge.H, 273.15 + 25, self.T_regeneration)
    
    def _cost(self):
        design_results = self.design_results
        baseline_purchase_costs = self.baseline_purchase_costs
        baseline_purchase_costs.update(self._vessel_purchase_cost(
            design_results['Weight'], design_results['Diameter'], design_results['Length']))
        for i, j in baseline_purchase_costs.items():
            baseline_purchase_costs[i] *= design_results['Number of reactors']


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
from setuptools import setup

setup(
    name='biosteam',
    packages=['biosteam'],
    license='MIT',
    version='2.31.23',
    description='The Biorefinery Simulation and Techno-Economic Analysis Modules',
    long_description=open('README.rst', encoding='utf-8').read(),
    author='Yoel Cortes-Pena',
    install_requires=['IPython>=7.9.0',
                      'thermosteam>=0.27.18', 
                      'graphviz>=0.17',
                      'chaospy>=3.3.9'],
    extras_require={ 
        'dev': [
            'biorefineries>=2.23.10',
            'sympy',
            'sphinx', 
            'sphinx_rtd_theme', 
            'pyyaml',
            'pytest-cov',
            'coveralls',
            'SALib',
        ]
    }, 
    package_data=
        {'biosteam': ['report/*',
                      'digraph/*',
                      'utils/*',
                      'evaluation/*', 
                      'evaluation/evaluation_tools/*',
                      'process_tools/*',
                      'plots/*',
                      'units/*',
                      'units/facilities/hxn/*',
                      'units/design_tools/*',
                      'units/facilities/*',
                      'units/decorators/*',
                      ]},
    platforms=['Windows', 'Mac', 'Linux'],
    author_email='yoelcortes@gmail.com',
    url='https://github.com/BioSTEAMDevelopmentGroup/biosteam',
    download_url='https://github.com/BioSTEAMDevelopmentGroup/biosteam.git',
    classifiers=['License :: OSI Approved :: University of Illinois/NCSA Open Source License',
                 'Development Status :: 3 - Alpha',
                 'Environment :: Console',
                 'Topic :: Scientific/Engineering',
                 'Topic :: Scientific/Engineering :: Chemistry',
                 'Topic :: Scientific/Engineering :: Mathematics',
                 'Intended Audience :: Developers',
                 'Intended Audience :: Education',
                 'Intended Audience :: Manufacturing',
                 'Intended Audience :: Science/Research',
                 'Natural Language :: English',
                 'Operating System :: MacOS',
                 'Operating System :: Microsoft :: Windows',
                 'Operating System :: POSIX',
                 'Operating System :: POSIX :: BSD',
                 'Operating System :: POSIX :: Linux',
                 'Operating System :: Unix',
                 'Programming Language :: Python :: 3.6',
                 'Programming Language :: Python :: 3.7',
                 'Programming Language :: Python :: 3.8',
                 'Programming Language :: Python :: Implementation :: CPython',
                 'Topic :: Education'],
    keywords=['chemical process simmulation', 'bioprocess engineering', 'mass and energy balance', 'material properties', 'phase equilibrium', 'CABBI', 'biorefinery', 'biofuel', 'bioproducts'],
)


# -*- coding: utf-8 -*-
"""
Created on Sat Dec  5 13:25:23 2020

@author: yrc2
"""
import pytest
import biosteam as bst
import os

folder = os.path.dirname(__file__)
folder = os.path.join(folder, 'Diagram Sources')
DISPLAY = False

def save_diagrams():
    import biorefineries.sugarcane as sc
    f = sc.C201.diagram(display=DISPLAY)
    file = os.path.join(folder, 'Clarifier.txt')
    expected_source = open(file, 'w')
    expected_source.write(f.source)
    for kind in ('thorough', 'cluster', 'surface', 'minimal'):
        f = sc.sugarcane_sys.diagram(kind, display=DISPLAY)
        file = os.path.join(folder, f'sugarcane {kind}.txt')
        expected_source = open(file, 'w')
        expected_source.write(f.source)
        expected_source.close()

def test_unit_diagram():
    import biorefineries.sugarcane as sc
    sc.load() # Reload system to make sure all is consistent
    f = sc.C201.diagram(display=DISPLAY)
    file = os.path.join(folder, 'Clarifier.txt')
    with open(file) as f_expected:
        expected_source = f_expected.read()
        assert set(f.source.split()) == set(expected_source.split())
    bst.process_tools.default()

def test_system_thorough_diagram():
    import biorefineries.sugarcane as sc
    f = sc.sugarcane_sys.diagram('thorough', display=DISPLAY)
    file = os.path.join(folder, 'sugarcane thorough.txt')
    with open(file) as f_expected:
        expected_source = f_expected.read()
        assert set(f.source.split()) == set(expected_source.split())
    bst.process_tools.default()

def test_system_cluster_diagram():
    import biorefineries.sugarcane as sc
    f = sc.sugarcane_sys.diagram('cluster', display=DISPLAY)
    file = os.path.join(folder, 'sugarcane cluster.txt')
    with open(file) as f_expected:
        expected_source = f_expected.read()
        assert set(f.source.split()) == set(expected_source.split())
    bst.process_tools.default()
    
# TODO: Find out why this test is not working
# def test_system_surface_diagram():
#     import biorefineries.sugarcane as sc
#     f = sc.sugarcane_sys.diagram('surface', display=DISPLAY)
#     file = os.path.join(folder, 'sugarcane surface.txt')
#     with open(file) as f_expected:
#         expected_source = f_expected.read()
#         assert set(f.source.split()) == set(expected_source.split())
#     bst.process_tools.default()
    
# def test_system_minimal_diagram():
#     import biorefineries.sugarcane as sc
#     f = sc.sugarcane_sys.diagram('minimal', display=DISPLAY)
#     file = os.path.join(folder, 'sugarcane minimal.txt')
#     with open(file) as f_expected:
#         expected_source = f_expected.read()
#         assert set(f.source.split()) == set(expected_source.split())
#     bst.process_tools.default()
    
if __name__ == '__main__':
    test_unit_diagram()
    # test_system_surface_diagram()
    test_system_thorough_diagram()
    test_system_cluster_diagram()
    # test_system_minimal_diagram()


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from thermosteam.units_of_measure import (
    convert, DisplayUnits, AbsoluteUnitsOfMeasure, get_dimensionality,
    heat_utility_units_of_measure
)
from thermosteam.utils import unregistered, units_of_measure
from thermosteam import Thermo, Stream, ThermalCondition, settings
from .exceptions import DimensionError
from math import copysign

__all__ = ('HeatUtility', 'UtilityAgent')

# Costs from Table 17.1 in Warren D.  Seider et al.-Product and Process Design Principles_ Synthesis, Analysis and Evaluation-Wiley (2016)
# ^This table was made using data from Busche, 1995
# Entry temperature conditions of coolants taken from Table 12.1 in Warren, 2016

mol_basis_units = AbsoluteUnitsOfMeasure('kmol')
mass_basis_units = AbsoluteUnitsOfMeasure('kg')
energy_basis_units = AbsoluteUnitsOfMeasure('kJ')

# %% Utility agents

@unregistered
class UtilityAgent(Stream):
    """
    Create a UtilityAgent object that defines a utility option.
    
    Parameters
    ----------
    ID='' : str
        A unique identification. If ID is None, utility agent will not be registered.
        If no ID is given, utility agent will be registered with a unique ID.
    flow=() : tuple
        All flow rates corresponding to chemical `IDs`.
    phase='l' : 'l', 'g', or 's'
        Either gas (g), liquid (l), or solid (s).
    T=298.15 : float
        Temperature [K].
    P=101325 : float
        Pressure [Pa].
    units='kmol/hr' : str
        Flow rate units of measure (only mass, molar, and
        volumetric flow rates are valid).
    thermo=None : Thermo
        Thermodynamic equilibrium package. Defaults to `biosteam.settings.get_thermo()`.
    T_limit : float, optional
        Temperature limit of outlet utility streams [K]. If no limit is given,
        phase change is assumed. If utility agent heats up, `T_limit` is
        the maximum temperature. If utility agent cools down, `T_limit` is
        the minumum temperature. 
    heat_transfer_price : float
        Price of transfered heat [USD/kJ].
    regeneration_price : float
        Price of regenerating the fluid for reuse [USD/kmol].
    heat_transfer_efficiency : float
        Fraction of heat transfered accounting for losses to the environment (must be between 0 to 1).
    **chemical_flows : float
        ID - flow pairs.
        
    """
    __slots__ = ('T_limit', '_heat_transfer_price',
                 '_regeneration_price', 'heat_transfer_efficiency')
    def __init__(self, ID='', flow=(), phase='l', T=298.15, P=101325., units='kmol/hr',
                 thermo=None, T_limit=None, heat_transfer_price=0.0,
                 regeneration_price=0.0, heat_transfer_efficiency=1.0,                  
                 **chemical_flows):
        self._thermal_condition = ThermalCondition(T, P)
        thermo = self._load_thermo(thermo)
        self._init_indexer(flow, phase, thermo.chemicals, chemical_flows)
        if units != 'kmol/hr':
            name, factor = self._get_flow_name_and_factor(units)
            flow = getattr(self, name)
            flow[:] = self.mol / factor
        self._islinked = False
        self._sink = self._source = None
        self.reset_cache()
        self._register(ID)
        self.T_limit = T_limit
        self.heat_transfer_price = heat_transfer_price
        self.regeneration_price = regeneration_price
        self.heat_transfer_efficiency = heat_transfer_efficiency
    
    def iscooling_agent(self):
        T_limit = self.T_limit 
        return T_limit > self.T if T_limit else self.phase == 'l'
    
    def isheating_agent(self):
        T_limit = self.T_limit 
        return T_limit < self.T if T_limit else self.phase == 'g'
    
    @property
    def price(self):
        raise AttributeError(f"'{type(self).__name__}' object has no attribute 'price'")
    @price.setter
    def price(self, price):
        raise AttributeError(f"'{type(self).__name__}' object has no attribute 'price'")
    
    @property
    def cost(self):
        raise AttributeError(f"'{type(self).__name__}' object has no attribute 'cost'")
    @cost.setter
    def cost(self, cost):
        raise AttributeError(f"'{type(self).__name__}' object has no attribute 'cost'")
    
    def to_stream(self, ID=None):
        """
        Return a copy as a :class:`~thermosteam.Stream` object.

        Examples
        --------
        >>> import biosteam as bst
        >>> bst.settings.set_thermo(['Water', 'Ethanol']) 
        >>> cooling_water = bst.HeatUtility.get_agent('cooling_water')
        >>> cooling_water_copy = cooling_water.to_stream('cooling_water_copy')
        >>> cooling_water_copy.show(flow='kg/hr')
        Stream: cooling_water_copy
         phase: 'l', T: 305.37 K, P: 101325 Pa
         flow (kg/hr): Water  18
        
        """
        new = Stream.__new__(Stream)
        new._islinked = False
        new._sink = new._source = None
        new._thermo = self._thermo
        new._imol = self._imol.copy()
        new._thermal_condition = self._thermal_condition.copy()
        new.reset_cache()
        new._price = 0.
        new.characterization_factors = {}
        new.ID = ID
        return new
    
    @property
    def heat_transfer_price(self):
        """Price of transfered heat [USD/kJ]."""
        return self._heat_transfer_price
    @heat_transfer_price.setter
    def heat_transfer_price(self, price):
        assert price >= 0, "heat transfer price cannot be negative"
        self._heat_transfer_price = price
    
    @property
    def regeneration_price(self):
        """Price of regenerating the fluid for reuse [USD/kmol]."""
        return self._regeneration_price
    @regeneration_price.setter
    def regeneration_price(self, price):
        assert price >= 0, "regeneration price cannot be negative"
        self._regeneration_price = price
    
    def _info_phaseTP(self, phase, T_units, P_units):
        T_limit = convert(self.T_limit, 'K', T_units)
        T_limit = f"{T_limit:.3g} K" if T_limit else "None"
        T = convert(self.T, 'K', T_units)
        P = convert(self.P, 'Pa', P_units)
        s = '' if isinstance(phase, str) else 's'
        ht_price = self.heat_transfer_price
        rg_price = self.regeneration_price
        ht_eff = self.heat_transfer_efficiency
        return (f" heat_transfer_efficiency: {ht_eff:.3f}\n"
                f" heat_transfer_price: {ht_price:.3g} USD/kJ\n"
                f" regeneration_price: {rg_price:.3g} USD/kmol\n"
                f" T_limit: {T_limit}\n"
                f" phase{s}: {repr(phase)}\n"
                f" T: {T:.5g} {T_units}\n"
                f" P: {P:.6g} {P_units}\n"
        )
    def __repr__(self):
        return f"<{type(self).__name__}: {self.ID}>"

       
# %%

@units_of_measure(heat_utility_units_of_measure)
class HeatUtility:
    """
    Create an HeatUtility object that can choose a utility stream and 
    calculate utility requirements. It can calculate required flow rate,
    temperature change, or phase change of utility. Calculations assume 
    counter current flow rate.
    
    Parameters
    ----------
    heat_transfer_efficiency=None : float, optional
        Enforced fraction of heat transfered from utility (due
        to losses to environment).
    
    Attributes
    ----------
    cooling_agents : list[UtilityAgent]
        All cooling utilities available.
    heating_agents : list[UtilityAgent]
        All heating utilities available.
    duty : float
        Total heat transfered from utility to both the process and the environment [kJ/hr].
    unit_duty : float
        Effective heat transfered from utility to the unit operation [kJ/hr].
    flow : float
        Flow rate of utility [kmol/hr].
    cost : float
        Cost of utility [USD/hr].
    heat_transfer_efficiency : float
        Enforced fraction of heat transfered from utility (due
        to losses to environment).
    T_pinch : float
        If cooling, the minimum utility temperature required.
        If heating, the maximum utility temperature required.
    iscooling : float
        Whether the utility is cooling the process.
    agent : UtilityAgent
        Utility agent being used.
    inlet_utility_stream : :class:`~thermosteam.Stream`
    outlet_utility_stream : :class:`~thermosteam.Stream`
    
    Examples
    --------
    Create a heat utility:
        
    .. code-block:: python
    
       >>> from biosteam import HeatUtility
       >>> hu = HeatUtility()
       >>> hu.show()
       HeatUtility: None
        duty: 0
        flow: 0
        cost: 0
    
    Calculate utility requirement by calling it with a duty (kJ/hr), and entrance and exit temperature (K):
        
    .. code-block:: python
    
       >>> hu(1000, 300, 350)
       >>> hu.show()
       HeatUtility: low_pressure_steam
        duty: 1.05e+03 kJ/hr
        flow: 0.0271 kmol/hr
        cost: 0.00645 USD/hr
   
    All results are accessible:
        
    .. code-block:: python
    
       >>> hu.ID, hu.duty, hu.flow, hu.cost
       ('low_pressure_steam', 1052., 0.02711, 0.006448)
           
    """
    __slots__ = ('inlet_utility_stream', 'outlet_utility_stream', 'duty',
                 'flow', 'cost', 'heat_transfer_efficiency', 'T_pinch',
                 'heat_exchanger', 'iscooling', 'agent', 'unit_duty')
    #: [float] Minimum approach temperature difference. Used to assign
    #: the pinch temperature of the utility stream.
    dT = 5  
    
    #: [DisplayUnits] Units of measure for IPython display.
    display_units = DisplayUnits(duty='kJ/hr', flow='kmol/hr', cost='USD/hr')

    #: [Thermo] Used broadly throughout BioSTEAM utilities.
    thermo_water = Thermo(['Water'])
    
    #: [Thermo] Used for BioSTEAM refrigeration utility.
    thermo_propane = Thermo(['Propane'])

    #: dict[tuple[str, str], tuple[float, AbsoluteUnitsOfMeasure]] 
    #: Characterization factor data (value and units) by agent ID and impact key.
    characterization_factors = {}
    
    @classmethod
    def set_CF(self, ID, key, value, basis=None, units=None):
        """
        Set the cacharacterization factor of a utility agent for a given impact 
        key.

        Parameters
        ----------
        ID : str
            ID of utility agent.
        key : str
            Key of impact indicator.
        value : float
            Characterization factor value.
        basis : str, optional
            Basis of characterization factor. Valid dimensions include weight, 
            molar, and energy (e.g. 'kg', 'kmol', 'kJ'). Defaults to 'kg'.
        units : str, optional
            Units of impact indicator. Before using this argument, the default units 
            of the impact indicator should be defined with 
            thermosteam.settings.define_impact_indicator.
            Units must also be dimensionally consistent with the default units.
            
        Raises
        ------
        ValueError
            When the duty characterization factor for a cooling agent is positive (should be negative).
        DimensionError
            When characterization factor is not given in dimensions of
            weight, molar, or energy.

        See Also
        --------
        HeatUtility.get_CF
        
        Examples
        --------
        Set the GWP characterization factor for low pressure steam at 
        88.44 kg CO2e / mmBtu (GREET 2020; Steam Production via Small Boiler from North American Natural Gas):
        
        >>> import biosteam as bst
        >>> bst.HeatUtility.set_CF('low_pressure_steam', 'GWP [kg CO2e]', 88.44, basis='MMBtu')
        
        Retrieve the GWP characterization factor for low pressure steam on a
        Btu basis:
        
        >>> bst.HeatUtility.get_CF('low_pressure_steam', 'GWP [kg CO2e]', basis='Btu')
        8.844e-05

        """
        agent = self.get_agent(ID)
        if units is not None:
            original_units = settings.get_impact_indicator_units(units)
            value = original_units.unconvert(value, units)
        if basis is None:
            basis_units = mass_basis_units
        else:
            dim = get_dimensionality(basis)
            if dim == mol_basis_units.dimensionality: 
                basis_units = mol_basis_units
            elif dim == mass_basis_units.dimensionality:
                basis_units = mass_basis_units
            elif dim == energy_basis_units.dimensionality:
                basis_units = energy_basis_units
                if agent.iscooling_agent() and value > 0.: 
                    raise ValueError(
                        'duty characterization factor must be negative for cooling agents'
                    )
            else:
                raise DimensionError(
                    "dimensions for characterization factors must be in a molar, "
                   f"mass or energy basis, not '{dim}'"
                )
            value *= basis_units.conversion_factor(basis)
        self.characterization_factors[agent.ID, key] = value, basis_units

    @classmethod
    def get_CF(self, ID, key, basis=None, units=None):
        """
        Return the characterization factor of a utility agent for a given impact 
        key.

        Parameters
        ----------
        ID : str
            ID of utility agent.
        key : str
            Key of impact indicator.
        basis : str, optional
            Basis of characterization factor. Valid dimensions include weight, 
            molar, and energy (e.g. 'kg', 'kmol', 'kJ'). Defaults to 'kg'.
        units : str, optional
            Units of impact indicator. Before using this argument, the default units 
            of the impact indicator should be defined with 
            thermosteam.settings.define_impact_indicator.
            Units must also be dimensionally consistent with the default units.
            
        Raises
        ------
        DimensionalityError
            When the characterization factor cannot be converted to the given basis 
            due to inconsistent dimensions with the original basis.

        See Also
        --------
        HeatUtility.set_CF

        """
        try:
            value, basis_units = self.characterization_factors[ID, key]
        except KeyError:
            if basis is None:
                return 0., None
            else:
                return 0.
        if units is not None:
            original_units = settings.get_impact_indicator_units(units)
            value = original_units.convert(value, units)
        if basis is None:
            return value, basis_units._units
        else:
            return value / basis_units.conversion_factor(basis)

    def get_impact(self, key):
        agent = self.agent
        CF, basis_units = self.get_CF(agent.ID, key)
        if CF == 0.: return 0.
        if basis_units == 'kg':
            return self.flow * agent.MW * CF
        elif basis_units == 'mol':
            return self.flow * CF
        elif basis_units == 'kJ':
            return self.duty * CF
        else:
            raise RuntimeError("unknown error")

    def get_inventory(self, key):
        agent = self.agent
        CF, basis_units = self.get_CF(agent.ID, key)
        if basis_units == 'kg':
            return self.flow * agent.MW, basis_units
        elif basis_units == 'mol':
            return self.flow, basis_units
        elif basis_units == 'kJ':
            return self.duty, basis_units
        else:
            raise RuntimeError("unknown error")

    def __init__(self, heat_transfer_efficiency=None, heat_exchanger=None):
        self.heat_transfer_efficiency = heat_transfer_efficiency
        self.heat_exchanger = heat_exchanger
        self.empty()

    def __bool__(self):
        return bool(self.agent)

    @property
    def ID(self):
        """[str] ID of utility agent being used."""
        agent = self.agent
        return agent.ID if agent else ""
    @ID.setter
    def ID(self, ID):
        """[str] ID of utility agent being used."""
        self.agent = self.get_agent(ID)

    @classmethod
    def default_agents(cls):
        """Reset all agents back to BioSTEAM's defaults."""
        cls.default_heating_agents()
        cls.default_cooling_agents()

    @classmethod
    def default_heating_agents(cls):
        """Reset all heating agents back to BioSTEAM's defaults."""
        thermo_water = cls.thermo_water
        low_pressure_steam = UtilityAgent(
            'low_pressure_steam',
            Water=1, T=412.189, P=344738.0, phase='g',
            thermo=thermo_water,
            regeneration_price = 0.2378,
            heat_transfer_efficiency = 0.95,
        )
        medium_pressure_steam = UtilityAgent(
            'medium_pressure_steam',
            Water=1, T=454.770, P=1.041e+6, phase='g',
            thermo=thermo_water,
            regeneration_price = 0.2756,
            heat_transfer_efficiency = 0.90,
        ) 
        high_pressure_steam = UtilityAgent(
            'high_pressure_steam',
            Water=1, T=508.991, P=3.11e+6, phase='g', 
            thermo=thermo_water,
            regeneration_price = 0.3171,
            heat_transfer_efficiency = 0.85,
        ) 
        cls.heating_agents = [low_pressure_steam, 
                              medium_pressure_steam, 
                              high_pressure_steam]
        
    @classmethod
    def default_cooling_agents(cls):
        """Reset all cooling agents back to BioSTEAM's defaults."""
        thermo_water = cls.thermo_water
        cooling_water = UtilityAgent(
            'cooling_water',
            Water=1, T=305.372, P=101325,
            thermo=thermo_water,
            T_limit = 324.817,
            regeneration_price = 4.8785e-4,
        )
        chilled_water = UtilityAgent(
            'chilled_water',
            Water=1, T=280.372, P=101325,
            thermo=thermo_water,
            T_limit = 300.372,
            heat_transfer_price = 5e-6
        )
        chilled_brine = UtilityAgent(
            'chilled_brine',
            Water=1, T=255.372, P=101325,
            thermo=thermo_water,
            T_limit = 275.372,
            heat_transfer_price = 8.145e-6,
        )
        propane = UtilityAgent(
            'propane',
            Propane=1,
            thermo=cls.thermo_propane,
            T=273.15 - 42.11,
            P=101325,
            heat_transfer_price = 13.17e-6,
            phase='l',
        )
        cls.cooling_agents = [cooling_water, 
                              chilled_water, 
                              chilled_brine,
                              propane]

    def copy(self):
        hu = HeatUtility()
        hu.copy_like(self)
        return hu

    def copy_like(self, other):
        """Copy all data from another heat utility."""
        if other.agent:
            self.load_agent(other.agent)
            self.inlet_utility_stream = other.inlet_utility_stream.copy()
            self.outlet_utility_stream = other.outlet_utility_stream.copy()
        self.flow = other.flow
        self.duty = other.duty
        self.unit_duty = other.unit_duty
        self.cost = other.cost
        self.heat_transfer_efficiency = other.heat_transfer_efficiency
        self.T_pinch = other.T_pinch
        self.iscooling = other.iscooling

    def scale(self, factor):
        """Scale utility data."""
        self.flow *= factor
        self.duty *= factor
        self.cost *= factor
        self.inlet_utility_stream.mol *= factor
        # No need to factor the outlet utility stream
        # because it shares the same flow rate data as the inlet

    def empty(self):
        """Remove utility requirements."""
        self.cost = self.flow = self.duty = self.unit_duty = 0
        self.iscooling = self.agent = self.T_pinch = None
        
    def set_utility_by_flow_rate(self, agent, F_mol):
        if F_mol == 0.: 
            self.empty()
            return
        self.load_agent(agent)
        heat_transfer_efficiency = self.heat_transfer_efficiency or agent.heat_transfer_efficiency
        if agent.T_limit: raise ValueError('agent must work by latent heat to set by flow rate')
        self.outlet_utility_stream.phase = 'g' if self.inlet_utility_stream.phase == 'l' else 'l'
        dH = self.outlet_utility_stream.Hvap
        self.duty = duty = dH * F_mol
        self.unit_duty = duty * heat_transfer_efficiency
        self.outlet_utility_stream.mol[:] = F_mol
        self.flow = F_mol
        self.cost = agent._heat_transfer_price * abs(duty) + agent._regeneration_price * F_mol
        
    def __call__(self, unit_duty, T_in, T_out=None, agent=None):
        """Calculate utility requirements given the essential parameters.
        
        Parameters
        ----------
        unit_duty : float
               Unit duty requirement (kJ/hr)
        T_in : float
               Inlet process stream temperature (K)
        T_out : float, optional
               Outlet process stream temperature (K)
        agent : UtilityAgent, optional
                Utility agent to use. Defaults to a suitable agent from 
                predefined heating/cooling utility agents.
        
        """
        if unit_duty == 0:
            self.empty()
            return
        T_out = T_out or T_in
        iscooling = unit_duty < 0.
        
        # Note: These are pinch temperatures at the utility inlet and outlet. 
        # Not to be confused with the inlet and outlet of the process stream.
        T_pinch_in, T_pinch_out = self.get_inlet_and_outlet_pinch_temperature(
            iscooling, T_in, T_out)

        ## Select heat transfer agent ##
        if agent:
            self.load_agent(agent)
        elif self.iscooling == iscooling and self.T_pinch == T_pinch_in:
            agent = self.agent
        else:
            self.iscooling = iscooling
            self.T_pinch = T_pinch_in
            if iscooling:
                agent = self.get_suitable_cooling_agent(T_pinch_in)
            else:
                agent = self.get_suitable_heating_agent(T_pinch_in)
            self.load_agent(agent)
            
        ## Calculate utility requirement ##
        heat_transfer_efficiency = self.heat_transfer_efficiency or agent.heat_transfer_efficiency
        duty = unit_duty/heat_transfer_efficiency
        if agent.T_limit:
            # Temperature change
            self.outlet_utility_stream.T = self.get_outlet_temperature(
                T_pinch_out, agent.T_limit, iscooling)
            dH = self.inlet_utility_stream.H - self.outlet_utility_stream.H
        else:
            # Phase change
            if self.inlet_utility_stream.phase == 'l' :
                self.outlet_utility_stream.phase = 'g'
                dH = -self.outlet_utility_stream.Hvap
            else:
                dH = self.outlet_utility_stream.Hvap
                self.outlet_utility_stream.phase = 'l'
        
        # Update utility flow
        self.outlet_utility_stream.mol[:] *= duty / dH
        
        # Update results
        self.unit_duty = unit_duty
        self.flow = F_mol = self.inlet_utility_stream.F_mol
        self.duty = duty
        self.cost = agent._heat_transfer_price * abs(duty) + agent._regeneration_price * F_mol

    @property
    def inlet_process_stream(self):
        """[:class`~thermosteam.Stream`] If a heat exchanger is available,
        this stream is the inlet process stream to the heat exchanger."""
        heat_exchanger = self.heat_exchanger
        if heat_exchanger:
            return heat_exchanger.ins[0]
        else:
            raise AttributeError('no heat exchanger available '
                                 'to retrieve process stream')

    @property
    def outlet_process_stream(self):
        """[:class`~thermosteam.Stream`] If a heat exchanger is available,
        this stream is the outlet process stream to the heat exchanger."""
        heat_exchanger = self.heat_exchanger
        if heat_exchanger:
            return heat_exchanger.outs[0]
        else:
            raise AttributeError('no heat exchanger available '
                                 'to retrieve process stream')
    
    @staticmethod
    def heat_utilities_by_agent(heat_utilities):
        """Return a dictionary of heat utilities sorted by agent."""
        heat_utilities = [i for i in heat_utilities if i.agent]
        heat_utilities_by_agent = {i.agent: [] for i in heat_utilities}
        for i in heat_utilities:
            heat_utilities_by_agent[i.agent].append(i)
        return heat_utilities_by_agent

    @classmethod
    def sum(cls, heat_utilities):
        """
        Return a HeatUtility object that reflects the sum of heat
        utilities.
        """
        heat_utility = cls()
        heat_utility.mix_from(heat_utilities)
        return heat_utility    

    @classmethod
    def sum_by_agent(cls, heat_utilities, agent=None):
        """
        Return a tuple of heat utilities that reflect the sum of heat utilities
        by agent. If an agent is specified, return a HeatUtility object that
        reflects the sum of all heat utilities with the given agent.
        
        """
        heat_utilities_by_agent = cls.heat_utilities_by_agent(heat_utilities)
        if agent:
            return cls.sum(heat_utilities_by_agent[agent])
        else:
            return tuple([cls.sum(i) for i in heat_utilities_by_agent.values()])

    @classmethod
    def get_agent(cls, ID):
        """Return utility agent with given ID."""
        for agent in cls.heating_agents + cls.cooling_agents:
            if agent.ID == ID: return agent
        raise LookupError(ID)

    @classmethod
    def get_heating_agent(cls, ID):
        """Return heating agent with given ID."""
        for agent in cls.heating_agents:
            if agent.ID == ID: return agent
        raise LookupError(ID)
  
    @classmethod
    def get_cooling_agent(cls, ID):
        """Return cooling agent with given ID."""
        for agent in cls.cooling_agents:
            if agent.ID == ID: return agent
        raise LookupError(ID)
    
    @classmethod
    def get_suitable_heating_agent(cls, T_pinch):
        """
        Return a heating agent that works at the pinch temperature.
        
        Parameters
        ----------
        T_pinch : float
            Pinch temperature [K].
        
        """
        for agent in cls.heating_agents:
            if T_pinch < agent.T: return agent
        raise RuntimeError(f'no heating agent that can heat over {T_pinch} K')    

    @classmethod
    def get_suitable_cooling_agent(cls, T_pinch):
        """Return a cooling agent that works at the pinch temperature.
        
        Parameters
        ----------
        T_pinch : float
            Pinch temperature [K].
        
        """
        for agent in cls.cooling_agents:
            if T_pinch > agent.T: return agent
        raise RuntimeError(f'no cooling agent that can cool under {T_pinch} K')    

    def load_agent(self, agent):
        """Initialize utility streams with given agent."""
        # Initialize streams
        self.inlet_utility_stream = agent.to_stream()
        self.outlet_utility_stream = self.inlet_utility_stream.flow_proxy()
        self.agent = agent

    def mix_from(self, heat_utilities):
        """Mix all heat utilities to this heat utility."""
        heat_utilities = [i for i in heat_utilities if i.agent]
        N_heat_utilities = len(heat_utilities)
        if N_heat_utilities == 0:
            self.empty()
        elif N_heat_utilities == 1:
            self.copy_like(heat_utilities[0])
        else:
            heat_utility, *other_heat_utilities = heat_utilities
            agent = heat_utility.agent
            ID = agent.ID
            assert all([i.agent.ID == ID for i in other_heat_utilities]), (
                "utility agent must be the same to mix heat utilities")
            self.load_agent(agent)
            self.flow = self.inlet_utility_stream.F_mol = sum([i.flow for i in heat_utilities])
            self.duty = sum([i.duty for i in heat_utilities])
            self.unit_duty = sum([i.unit_duty for i in heat_utilities])
            self.cost = sum([i.cost for i in heat_utilities])
            self.heat_transfer_efficiency = None
            self.T_pinch = None
            self.iscooling = None

    def reverse(self):
        """Reverse direction of utility. If utility is being consumed,
        the utility is produced instead, and vice-versa."""
        self.flow *= -1
        self.duty *= -1
        self.unit_duty *= -1
        self.cost *= -1
        if self.duty:
            self.inlet_utility_stream, self.outlet_utility_stream = self.outlet_utility_stream, self.inlet_utility_stream

    # Subcalculations

    @staticmethod
    def get_outlet_temperature(T_pinch, T_limit, iscooling):
        """
        Return outlet temperature of the utility in a counter current heat exchanger

        Parameters
        ----------
        T_pinch : float
                  Pinch temperature of utility stream [K].
        iscooling : bool
                  True if utility is loosing energy.

        """
        if iscooling:
            return T_limit if T_limit and T_limit < T_pinch else T_pinch
        else:
            return T_limit if T_limit and T_limit > T_pinch else T_pinch
        
    @classmethod
    def get_inlet_and_outlet_pinch_temperature(cls, iscooling, T_in, T_out):
        """Return pinch inlet and outlet temperature of utility."""
        dT = cls.dT
        if iscooling:
            if T_in + 1e-1 < T_out:
                raise ValueError("inlet must be hotter than outlet if cooling")
            T_pinch_in = T_out - dT
            T_pinch_out = T_in - dT
        else:
            if T_in > T_out + 1e-1:
                raise ValueError("inlet must be cooler than outlet if heating")
            T_pinch_in = T_out + dT
            T_pinch_out = T_in + dT
        return T_pinch_in, T_pinch_out

    def _info_data(self, duty, flow, cost):
        # Get units of measure
        su = self.display_units
        duty_units = duty or su.duty
        flow_units = flow or su.flow
        cost_units = cost or su.cost
        
        # Change units and return info string
        flow = self.inlet_utility_stream.get_total_flow(flow_units)
        duty = convert(self.duty, 'kJ/hr', duty_units)
        cost = convert(self.cost, 'USD/hr', cost_units)
        return duty, copysign(flow, self.flow), cost, duty_units, flow_units, cost_units
        
    def __repr__(self):
        if self.agent:
            duty, flow, cost, duty_units, flow_units, cost_units = self._info_data(None, None, None)
            return f'<{self.ID}: {self.duty:.3g} {duty_units}, {self.flow:.3g} {flow_units}, {self.cost:.3g} {cost_units}>'
        else:
            return f'<{type(self).__name__}: None>'
        
    # Representation
    def _info(self, duty, flow, cost):
        """Return string related to specifications"""
        if not self.agent:
            return (f'{type(self).__name__}: None\n'
                    +' duty: 0\n'
                    +' flow: 0\n'
                    +' cost: 0')
        else:
            (duty, flow, cost, duty_units,
             flow_units, cost_units) = self._info_data(duty, flow, cost)
            return (f'{type(self).__name__}: {self.ID}\n'
                    +f' duty:{duty: .3g} {duty_units}\n'
                    +f' flow:{flow: .3g} {flow_units}\n'
                    +f' cost:{cost: .3g} {cost_units}')
            

    def show(self, duty=None, flow=None, cost=None):
        """Print all specifications"""
        print(self._info(duty, flow, cost))
    _ipython_display_ = show

HeatUtility.default_agents()


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2021-2022, Yalin Li <zoe.yalin.li@gmail.com>,
#                          Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""

def test_qsdsan():
    from exposan import bwaise as bw
    # TODO: Make tests with assertions
    # Just make sure it can simulate for now
    for i in (bw.sysA, bw.sysB, bw.sysC): i.simulate() 

if __name__ == '__main__':
    test_qsdsan() 


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from .. import Unit, PowerUtility
from thermosteam import MultiStream, separations
from math import pi
import numpy as np
from . import design_tools as design
from .splitting import Splitter
from .heat_exchange import HX, HXutility
from .._graphics import vertical_vessel_graphics
import biosteam as bst

exp = np.exp
ln = np.log

# USDA Biodiesel: 
#V = np.pi*D**2/4*Ht*0.3048**3
#if not (0.1 < V < 70):
#    raise DesignError(f"Volume is out of bounds for costing")
#lambda V, CE: CE*13*V**0.62 # V (m3)
__all__ = ('Flash', 'SplitFlash', 'RatioFlash')


# %% Flash

class Flash(design.PressureVessel, Unit):
    """
    Create an equlibrium based flash drum with the option of having light
    non-keys and heavy non-keys completly separate into their respective
    phases. Design procedure is based on heuristics by Wayne D. Monnery & 
    William Y. Svrcek [1]_. Purchase costs are based on correlations by
    Mulet et al. [2]_ [3]_ as compiled by Warren et. al. [4]_.

    Parameters
    ----------
    ins : stream
        Inlet fluid.
    outs : stream sequence
        * [0] Vapor product
        * [1] Liquid product
    P=None : float
        Operating pressure [Pa].
    Q=None : float
        Duty [kJ/hr].
    T=None : float
        Operating temperature [K].
    V=None : float
        Molar vapor fraction.
    x=None : float
        Molar composition of liquid (for binary mixtures).
    y=None : float
        Molar composition of vapor (for binary mixtures).
    vessel_material : str, optional
        Vessel construction material. Defaults to 'Carbon steel'.
    vacuum_system_preference : 'Liquid-ring pump', 'Steam-jet ejector', or 'Dry-vacuum pump'
        If a vacuum system is needed, it will choose one according to this
        preference. Defaults to 'Liquid-ring pump'.
    has_glycol_groups=False : bool
        True if glycol groups are present in the mixture.
    has_amine_groups=False : bool
        True if amine groups are present in the mixture.
    vessel_type=None : 'Horizontal' or 'Vertical', optional
        Vessel separation type. If not specified, the vessel type will be chosen
        according to heuristics.
    holdup_time=15.0 : float
        Time it takes to raise liquid to half full [min].
    surge_time=7.5 : float
        Time it takes to reach from normal to maximum liquied level [min].
    has_mist_eliminator : bool
        True if using a mist eliminator pad.

    Notes
    -----
    You may only specify two of the following parameters: P, Q, T, V, x, and y.
    Additionally, If x or y is specified, the other parameter must be either
    P or T (e.g., x and V is invalid).

    Examples
    --------
    >>> from biosteam.units import Flash
    >>> from biosteam import Stream, settings
    >>> settings.set_thermo(['Water', 'Glycerol'], cache=True)
    >>> feed = Stream('feed', Glycerol=300, Water=1000)
    >>> bp = feed.bubble_point_at_P() # Feed at bubble point T
    >>> feed.T = bp.T
    >>> F1 = Flash('F1',
    ...            ins=feed,
    ...            outs=('vapor', 'crude_glycerin'),
    ...            P=101325, # Pa
    ...            T=410.15) # K
    >>> F1.simulate()
    >>> F1.show(T='degC', P='atm')
    Flash: F1
    ins...
    [0] feed
        phase: 'l', T: 100.7 degC, P: 1 atm
        flow (kmol/hr): Water     1e+03
                        Glycerol  300
    outs...
    [0] vapor
        phase: 'g', T: 137 degC, P: 1 atm
        flow (kmol/hr): Water     958
                        Glycerol  2.32
    [1] crude_glycerin
        phase: 'l', T: 137 degC, P: 1 atm
        flow (kmol/hr): Water     42.4
                        Glycerol  298
    >>> F1.results()
    Flash                                                   Units            F1
    Medium pressure steam Duty                              kJ/hr      4.82e+07
                          Flow                            kmol/hr      1.33e+03
                          Cost                             USD/hr           367
    Design                Vessel type                                  Vertical
                          Length                               ft          15.5
                          Diameter                             ft           8.5
                          Weight                               lb      9.57e+03
                          Wall thickness                       in         0.438
                          Vessel material                          Carbon steel
    Purchase cost         Heat_exchanger - Floating head      USD      4.24e+04
                          Vertical pressure vessel            USD      4.63e+04
                          Platform and ladders                USD      1.39e+04
    Total purchase cost                                       USD      1.03e+05
    Utility cost                                           USD/hr           367


    References
    ----------
    .. [1] "Design Two-Phase Separators Within the Right Limits", Chemical
        Engineering Progress Oct, 1993.

    .. [2] Mulet, A., A. B. Corripio, and L. B. Evans, “Estimate Costs of
        Pressure Vessels via Correlations,” Chem. Eng., 88(20), 145–150 (1981a).

    .. [3] Mulet, A., A.B. Corripio, and L.B.Evans, “Estimate Costs of
        Distillation and Absorption Towers via Correlations,” Chem. Eng., 88(26), 77–82 (1981b).

    .. [4] Seider, W. D., Lewin,  D. R., Seader, J. D., Widagdo, S., Gani,
        R., & Ng, M. K. (2017). Product and Process Design Principles. Wiley.
        Cost Accounting and Capital Cost Estimation (Chapter 16)
    
    """
    auxiliary_unit_names = ('heat_exchanger',)
    _units = {'Length': 'ft',
              'Diameter': 'ft',
              'Weight': 'lb',
              'Wall thickness': 'in'}
    _max_agile_design = (
        'Length',
        'Diameter',
        'Weight',
        'Wall thickness',
    )
    _F_BM_default = {'Liquid-ring pump': 1.0,
                     **design.PressureVessel._F_BM_default}
    _graphics = vertical_vessel_graphics 
    _N_outs = 2
    _N_heat_utilities = 0

    def __init__(self, ID='', ins=None, outs=(), thermo=None, *,
                 V=None, T=None, Q=None, P=None, y=None, x=None,
                 vessel_material='Carbon steel',
                 vacuum_system_preference='Liquid-ring pump',
                 has_glycol_groups=False,
                 has_amine_groups=False,
                 vessel_type=None,
                 holdup_time=15,
                 surge_time=7.5,
                 has_mist_eliminator=False):
        Unit.__init__(self, ID, ins, outs, thermo)
        self._load_components()
        
        #: Enforced molar vapor fraction
        self.V = V
        
        #: Enforced operating temperature (K)
        self.T = T
        
        #: [array_like] Molar composition of vapor (for binary mixture)
        self.y = y
        
        #: [array_like] Molar composition of liquid (for binary mixture)
        self.x = x
        
        #: Enforced duty (kJ/hr)
        self.Q = Q
        
        #: Operating pressure (Pa)
        self.P = P
        
        #: [str] Vessel construction material
        self.vessel_material = vessel_material

        #: [str] If a vacuum system is needed, it will choose one according to this preference.
        self.vacuum_system_preference = vacuum_system_preference
        
        #: [bool] True if glycol groups are present in the mixture
        self.has_glycol_groups = has_glycol_groups
        
        #: [bool] True if amine groups are present in the mixture
        self.has_amine_groups = has_amine_groups
        
        #: [str] 'Horizontal', 'Vertical', or 'Default'
        self.vessel_type = vessel_type
        
        #: [float] Time it takes to raise liquid to half full (min)
        self.holdup_time = holdup_time
        
        #: [float] Time it takes to reach from normal to maximum liquied level (min)
        self.surge_time = surge_time
        
        #: [bool] True if using a mist eliminator pad
        self.has_mist_eliminator = has_mist_eliminator
        
    def _load_components(self):
        self._multi_stream = ms = MultiStream(None, thermo=self.thermo)
        self.heat_exchanger = hx = HXutility(None, (None,), ms, thermo=self.thermo) 
        hx.owner = self.owner
        self.heat_utilities = (*hx.heat_utilities, bst.HeatUtility(), bst.HeatUtility())
        
    def reset_cache(self):
        self._multi_stream.reset_cache()
        self.heat_exchanger.reset_cache()
        
    @property
    def P(self):
        """Operating pressure (Pa)."""
        return self._P
    @P.setter
    def P(self, P):
        if P and P < 101325 and not self.power_utility:
            self.power_utility = PowerUtility()
        self._P = P

    def _default_vessel_type(self):
        vap, liq = self.outs
        F_mass_vap = vap.F_mass
        F_mass_liq = liq.F_mass 
        assert F_mass_liq, "no liquid effluent; flash vessel must have liquid to default vessel type"
        return 'Vertical'if F_mass_vap / F_mass_liq > 0.1 else 'Horizontal'

    def _run(self):
        separations.vle(self.ins[0], *self.outs, self.T, self.P, self.V, 
                        self.Q, self.x, self.y, self._multi_stream)
            
    def _design(self):
        vessel_type = self.vessel_type
        if vessel_type == 'Vertical': 
            args = self._vertical_vessel_pressure_diameter_and_length()
        elif vessel_type == 'Horizontal': 
            args = self._horizontal_vessel_pressure_diameter_and_length()
        else: raise RuntimeError('unknown vessel type') # pragma: no cover
        feed = self.heat_exchanger.ins[0]
        feed.mix_from(self.ins)
        feed.vle(P=self.outs[0].P, H=feed.H)
        if self.Q == 0:
            self.heat_exchanger.baseline_purchase_costs.clear()
            self.heat_exchanger.purchase_costs.clear()
            self.heat_exchanger.installed_costs.clear()
        else:
            self.heat_exchanger._summary()
        self.design_results.update(
            self._vessel_design(*args)
        )

    def _cost(self):
        D = self.design_results
        self.baseline_purchase_costs.update(
            self._vessel_purchase_cost(D['Weight'], D['Diameter'], D['Length'])
        )
        self._cost_vacuum()

    def _cost_vacuum(self):
        P = self.P
        if not P or P > 101320: return 
        
        Design = self.design_results
        R = Design['Diameter'] * 0.5
        volume = 0.02832 * np.pi * Design['Length'] * R * R
        
        # If vapor is condensed, assume vacuum system is after condenser
        vapor = self.outs[0]
        hx = vapor.sink
        if isinstance(hx, HX):
            index = hx.ins.index(vapor)
            stream = hx.outs[index]
            if isinstance(stream, MultiStream): # pragma: no cover
                vapor = stream['g']
                F_mass = vapor.F_mass
                F_vol = vapor.F_vol
            else:
                if stream.phase == 'g': # pragma: no cover
                    F_mass = stream.F_mass
                    F_vol = stream.F_vol
                else:
                    F_mass = 0
                    F_vol = 0
        else:
            F_mass = vapor.F_mass
            F_vol = vapor.F_vol
        
        vacuum_results = design.compute_vacuum_system_power_and_cost(
            F_mass, F_vol, P, volume, self.vacuum_system_preference
        )
        self.baseline_purchase_costs['Vacuum system'] = vacuum_results['Cost']
        self.design_results['Vacuum system'] = vacuum_results['Name']
        hx_hu, vacuum_steam, vacuum_cooling_water = self.heat_utilities
        vacuum_steam.set_utility_by_flow_rate(vacuum_results['Heating agent'], vacuum_results['Steam flow rate'])
        if vacuum_results['Condenser']: 
            vacuum_cooling_water(-vacuum_steam.unit_duty, 373.15)
        else:
            vacuum_cooling_water.empty()
        self.power_utility(vacuum_results['Work'])

    def _design_parameters(self):
        # Retrieve run_args and properties
        vap, liq = self._outs
        rhov = vap.get_property('rho', 'lb/ft3')
        rhol = liq.get_property('rho', 'lb/ft3')
        P = liq.get_property('P', 'psi')  # Pressure (psi)

        vessel_type = self.vessel_type
        Th = self.holdup_time
        Ts = self.surge_time
        has_mist_eliminator = self.has_mist_eliminator

        # Calculate the volumetric flowrate
        Qv = vap.get_total_flow('ft^3 / s')
        Qll = liq.get_total_flow('ft^3 / min')

        # Calculate Ut and set Uv
        K = design.compute_Stokes_law_York_Demister_K_value(P)

        # Adjust K value
        if not has_mist_eliminator and vessel_type == 'Vertical': K /= 2

        # Adjust for amine or glycol groups:
        if self.has_glycol_groups: K *= 0.6
        elif self.has_amine_groups: K *= 0.8

        Ut = K*((rhol - rhov) / rhov)**0.5
        Uv = 0.75*Ut

        # Calculate Holdup and Surge volume
        Vh = Th*Qll
        Vs = Ts*Qll
        return rhov, rhol, P, Th, Ts, has_mist_eliminator, Qv, Qll, Ut, Uv, Vh, Vs

    def _vertical_vessel_pressure_diameter_and_length(self):
        rhov, rhol, P, Th, Ts, has_mist_eliminator, Qv, Qll, Ut, Uv, Vh, Vs = self._design_parameters()

        # Calculate internal diameter, Dvd
        Dvd = (4.0*Qv/(pi*Uv))**0.5
        if has_mist_eliminator:
            D = design.ceil_half_step(Dvd + 0.4)
        else:
            D = design.ceil_half_step(Dvd)

        # Obtaining low liquid level height, Hlll
        Hlll = 0.5
        if P < 300:
            Hlll = 1.25

        # Calculate the height from Hlll to Normal liq level, Hnll
        Hh = Vh/(pi/4.0*Dvd**2)
        if Hh < 1.0: Hh = 1.0

        # Calculate the height from Hnll to  High liq level, Hhll
        Hs = Vs/(pi/4.0*Dvd**2)
        if Hs < 0.5: Hs = 0.5

        # Calculate dN
        Qm = Qll + Qv
        lamda = Qll/Qm
        rhoM = rhol*lamda + rhov*(1-lamda)
        dN = (4*Qm/(pi*60.0/(rhoM**0.5)))**0.5
        dN = design.ceil_half_step(dN)

        # Calculate Hlin, assume with inlet diverter
        Hlin = 1.0 + dN

        # Calculate the vapor disengagement height
        Hv = 0.5*Dvd
        Hv2 = (2.0 if has_mist_eliminator else 3.0) + dN/2.0 # pragma: no cover
        if Hv2 < Hv: Hv = Hv2
        Hv = Hv

        # Calculate total height, Ht
        Hme = 1.5 if has_mist_eliminator else 0.0
        Ht = Hlll + Hh + Hs + Hlin + Hv + Hme
        Ht = design.ceil_half_step(Ht)

        # Find maximum and normal liquid level
        # Hhll = Hs + Hh + Hlll
        # Hnll = Hh + Hlll

        return P, D, Ht
        
    def _horizontal_vessel_pressure_diameter_and_length(self):
        rhov, rhol, P, Th, Ts, has_mist_eliminator, Qv, Qll, Ut, Uv, Vh, Vs = self._design_parameters()

        # Initialize LD
        if P > 0 and P <= 264.7:
            LD = 1.5/250.0*(P-14.7)+1.5
        elif P > 264.7 and P <= 514.7: # pragma: no cover
            LD = 1.0/250.0*(P-14.7)+2.0
        elif P > 514.7: # pragma: no cover
            LD = 5.0

        D = (4.0*(Vh+Vs)/(0.6*pi*LD))**(1.0/3.0)
        if D <= 4.0:
            D = 4.0
        else:
            D = design.ceil_half_step(D)

        for outerIter in range(50):
            At = pi*(D**2)/4.0 # Total area

            # Calculate Lower Liquid Area
            Hlll = round(0.5*D + 7.0)  
            Hlll = Hlll/12.0 # D is in ft but Hlll is in inches
            X = Hlll/D
            Y = design.HNATable(1, X)
            Alll = Y*At

            # Calculate the Vapor disengagement area, Av
            Hv = 0.2*D
            if has_mist_eliminator and Hv <= 2.0: Hv = 2.0
            elif Hv <= 1.0: Hv = 1.0
            else: Hv = design.ceil_half_step(Hv)
            Av = design.HNATable(1, Hv/D)*At
            
            # Calculate minimum length for surge and holdup
            L = (Vh + Vs)/(At - Av - Alll)
            # Calculate liquid dropout
            Phi = Hv/Uv
            # Calculate actual vapor velocity
            Uva = Qv/Av
            # Calculate minimum length for vapor disengagement
            Lmin = Uva*Phi
            Li = L
            
            for innerIter in range(50):
                if L < 0.8*Lmin: Hv += 0.5
                elif L > 1.2*Lmin:
                    if has_mist_eliminator and Hv <= 2.0: Hv = 2.0
                    elif not has_mist_eliminator and Hv <= 1.0: Hv = 1.0
                    else: Hv -= 0.5
                else: break
                Av = design.HNATable(1, Hv/D)*At
                Alll = design.HNATable(1, Hlll/D)*At
                Li = (Vh + Vs)/(At - Av - Alll)
                Phi = Hv/Uv
                Uva = Qv/Av
                Lmin = Uva*Phi
            
            L = Li
            LD = L/D
            # Check LD
            if LD < 1.2:
                if D <= 4.0: break
                else: D -= 0.5

            if LD > 7.2: # pragma: no cover
                D += 0.5
            else: break

        # Recalculate LD so it lies between 1.5 - 6.0
        while True:
            LD = L / D
            if (LD < 1.5) and D <= 4.0: L += 0.5
            elif LD < 1.5: D -= 0.5
            elif (LD > 6.0): D += 0.5
            else: break

        # # To check minimum Hv value
        # if int(has_mist_eliminator) == 1 and Hv <= 2.0:
        #     Hv = 2.0
        # if int(has_mist_eliminator) == 0 and Hv <= 1.0:
        #     Hv = 1.0

        # Calculate normal liquid level and High liquid level
        # Hhll = D - Hv
        # if (Hhll < 0.0):
        #     Hhll = 0.0
        # Anll = Alll + Vh/L
        # X = Anll/At
        # Y = HNATable(2, X)
        # Hnll = Y*D
        
        return P, D, L


# %% Special

class SplitFlash(Flash):
    line = 'Flash' 
    
    def __init__(self, ID='', ins=None, outs=(), thermo=None, *, split,
                 order=None, T=None, P=None, Q=None,
                 vessel_material='Carbon steel',
                 vacuum_system_preference='Liquid-ring pump',
                 has_glycol_groups=False,
                 has_amine_groups=False,
                 vessel_type=None,
                 holdup_time=15,
                 surge_time=7.5,
                 has_mist_eliminator=False):
        Splitter.__init__(self, ID, ins, outs, thermo, split=split, order=order)
        self._load_components()
        
        self.T = T #: Operating temperature (K)
        self.P = P #: Operating pressure (Pa)
        self.Q = Q #: Duty (kJ/hr)
        
        #: [str] Vessel construction material
        self.vessel_material = vessel_material

        #: [str] If a vacuum system is needed, it will choose one according to this preference.
        self.vacuum_system_preference = vacuum_system_preference
        
        #: [bool] True if glycol groups are present in the mixture
        self.has_glycol_groups = has_glycol_groups
        
        #: [bool] True if amine groups are present in the mixture
        self.has_amine_groups = has_amine_groups
        
        #: [str] 'Horizontal', 'Vertical', or 'Default'
        self.vessel_type = vessel_type
        
        #: [float] Time it takes to raise liquid to half full (min)
        self.holdup_time = holdup_time
        
        #: [float] Time it takes to reach from normal to maximum liquied level (min)
        self.surge_time = surge_time
        
        #: [bool] True if using a mist eliminator pad
        self.has_mist_eliminator = has_mist_eliminator
    
    split = Splitter.split
    V = None
    
    def _run(self):
        top, bot = self.outs
        feed, = self.ins
        feed_mol = feed.mol
        top.mol[:] = top_mol = feed_mol * self.split
        bot.mol[:] = feed_mol - top_mol
        top.phase = 'g'
        bot.phase = 'l'
        bot.T = top.T = self.T
        bot.P = top.P = self.P

    def _design(self):
        self.heat_exchanger.outs[0] = ms = self._multi_stream
        ms.mix_from(self.outs)
        super()._design()
    
# TODO: Remove this in favor of partition coefficients
class RatioFlash(Flash):
    _N_heat_utilities = 1

    def __init__(self, ID='', ins=None, outs=(), *,
                 K_chemicals, Ks, top_solvents=(), top_split=(),
                 bot_solvents=(), bot_split=()):
        Unit.__init__(self, ID, ins, outs)
        self._load_components()
        self.K_chemicals = K_chemicals
        self.Ks = Ks
        self.top_solvents = top_solvents
        self.top_split = top_split
        self.bot_solvents = bot_solvents
        self.bot_split = bot_split

    def _run(self):
        feed = self.ins[0]
        top, bot = self.outs
        indices = self.chemicals.get_index
        K_index = indices(self.K_chemicals)
        top_index = indices(self.top_solvents)
        bot_index = indices(self.bot_solvents)
        top_mol = top.mol; bot_mol = bot.mol; feed_mol = feed.mol
        top_mol[top_index] = feed_mol[top_index]*self.top_split
        bot_mol[top_index] = feed_mol[top_index]-top_mol[top_index]
        bot_mol[bot_index] = feed_mol[bot_index]*self.bot_split
        top_mol[bot_index] = feed_mol[bot_index]-bot_mol[bot_index]
        topnet = top_mol[top_index].sum()
        botnet = bot_mol[bot_index].sum()
        molnet = topnet+botnet
        top_mol[K_index] = self.Ks * topnet * feed_mol[K_index] / molnet  # solvent * mol ratio
        bot_mol[K_index] = feed_mol[K_index] - top_mol[K_index]
        top.T, top.P = feed.T, feed.P
        bot.T, bot.P = feed.T, feed.P

    def _design(self): # pragma: no cover
        self.heat_exchanger.outs[0] = ms = self._multi_stream
        ms.mix_from(self.outs)
        super()._design()

# %% Single Component

class Evaporator_PQ(Unit):
    _N_ins = 2
    _N_outs = 3
    @property
    def P(self):
        return self._P
    @P.setter
    def P(self, P):
        water = getattr(self.chemicals, '7732-18-5')
        self._T = T = water.Tsat(P)
        self._Hvap = water.Hvap(T)
        self._P = P
    @property
    def T(self):
        return self._T
    @T.setter
    def T(self, T): # pragma: no cover
        water = getattr(self.chemicals, '7732-18-5')
        self._P = water.Psat(T)
        self._Hvap = water.Hvap(T)
        self._T = T
    @property
    def V(self):
        return self._V
    
    def __init__(self, ID='', ins=None, outs=(), thermo=None, *, Q=0, P=101325):
        super().__init__(ID, ins, outs, thermo)
        self.Q = Q
        self.P = P
        self._V = None
    
    def _run(self):
        feed, utility_vapor = self.ins
        vapor, liquid, utility_liquid = self.outs
        
        # Optional if Q also comes from condensing a side stream
        Q = self.Q
        if not utility_vapor.isempty():
            utility_liquid.copy_like(utility_vapor)
            utility_liquid.phase = 'l'
            Q += utility_vapor.Hvap
        feed_H = feed.H
    
        # Set exit conditions
        vapor.T = liquid.T = self.T
        vapor.P = liquid.P = self.P
        liquid.phase = 'l'
        vapor.phase = 'g'
        liquid.copy_flow(feed)
        
        # Energy balance to find vapor fraction
        f = feed.imol['7732-18-5']
        H = feed_H + Q - liquid.H
        if f:
            V = H/(f * self._Hvap)
            if V < 0: 
                V = 0
            elif V > 1:
                V = 1
        else:
            V = 0
        evaporated = f * V
        vapor.imol['7732-18-5'] = evaporated
        liquid.imol['7732-18-5'] = (1-V)*f
        self.design_results['Heat transfer'] = Q
        self._V = V


class Evaporator_PV(Unit):
    _N_heat_utilities = 0
    _N_outs = 2
    
    @property
    def P(self):
        return self._P
    @P.setter
    def P(self, P):
        water = getattr(self.chemicals, '7732-18-5')
        self._T = water.Tsat(P)
        self._P = P
    @property
    def T(self):
        return self._T
    @T.setter
    def T(self, T): # pragma: no cover
        water = getattr(self.chemicals, '7732-18-5')
        self._P = water.Psat(T)
        self._T = T
    
    def __init__(self, ID='', ins=None, outs=(), thermo=None, *, V=0.5, P=101325):
        super().__init__(ID, ins, outs, thermo)
        self._multi_stream = ms = MultiStream(None, thermo=self.thermo)
        self.heat_exchanger = hx = HXutility(None, None, ms, thermo=self.thermo) 
        self.heat_utilities = hx.heat_utilities
        self.V = V
        self.P = P
        hx._ins = self._ins

    def _run(self):
        feed = self.ins[0]
        vapor, liquid = self.outs
        vapor.phase = 'g'
        vapor.T = liquid.T = self.T
        vapor.P = liquid.P = self.P
        H2O_index = self.chemicals.index('7732-18-5')
        water_mol = feed.mol[H2O_index]
        vapor.mol[H2O_index] = self.V * water_mol
        liquid_mol = liquid.mol
        liquid_mol[:] = feed.mol
        liquid_mol[H2O_index] = (1-self.V) * water_mol
        ms = self._multi_stream
        self.design_results['Heat transfer'] = self.heat_exchanger.Q = (vapor.H + liquid.H) - feed.H
        ms['g'].copy_like(vapor)
        ms['l'].copy_like(liquid)